#ifndef AMDIS_LAGRANGEINTERPOLRESTRICT_H
#define AMDIS_LAGRANGEINTERPOLRESTRICT_H

#include "InterpolRestrictMatrix.h"

namespace AMDiS {

  /** \brief
   * Defines the interpolation and restriction matrix for 1d-elements
   * and 1d-lagrange basis functions.
   */
  class IM_1d_1 : public InterpolRestrictMatrix
  {
  public:
    IM_1d_1() : 
      InterpolRestrictMatrix(2, 3, 2, false) 
    {}

  protected:
    void fillMemory()
    {
      directCopyDOFs_.resize(2);
      directCopyDOFs_[0] = 0;
      directCopyDOFs_[1] = 1;

      uniqueRowMatrix_[2][0] = 0.5;
      uniqueRowMatrix_[2][1] = 0.5;

      rowDOFPtrs_[0] = parentDOFs_ + 0;
      rowDOFPtrs_[1] = parentDOFs_ + 1;
      rowDOFPtrs_[2] = child0DOFs_ + 1;

      colDOFPtrs_[0] = parentDOFs_ + 0;
      colDOFPtrs_[1] = parentDOFs_ + 1;
    }
  };


  /** \brief
   * Defines the interpolation and restriction matrix for 1d-elements
   * and 2d-lagrange basis functions.
   */
  class IM_1d_2 : public InterpolRestrictMatrix
  {
  public:
    IM_1d_2() : 
      InterpolRestrictMatrix(3, 5, 3, false) 
    {}

  protected:
    void fillMemory()
    {
      directCopyDOFs_.push_back(0);
      directCopyDOFs_.push_back(1);

      uniqueRowMatrix_[2][2] = 1.0;

      uniqueRowMatrix_[3][0] = 0.375;
      uniqueRowMatrix_[3][1] = -0.125;
      uniqueRowMatrix_[3][2] = 0.75;

      uniqueRowMatrix_[4][0] = -0.125;
      uniqueRowMatrix_[4][1] = 0.375;
      uniqueRowMatrix_[4][2] = 0.75;

      rowDOFPtrs_[0] = parentDOFs_ + 0;
      rowDOFPtrs_[1] = parentDOFs_ + 1;
      rowDOFPtrs_[2] = child0DOFs_ + 1;
      rowDOFPtrs_[3] = child0DOFs_ + 2;
      rowDOFPtrs_[4] = child1DOFs_ + 2;

      colDOFPtrs_[0] = parentDOFs_ + 0;
      colDOFPtrs_[1] = parentDOFs_ + 1;
      colDOFPtrs_[2] = parentDOFs_ + 2;
    }
  };


  /** \brief
   * Defines the interpolation and restriction matrix for 1d-elements
   * and 4d-lagrange basis functions.
   */
  class IM_1d_4 : public InterpolRestrictMatrix
  {
  public:
    IM_1d_4() : 
      InterpolRestrictMatrix(5, 9, 5, false) 
    {}

  protected:
    void fillMemory()
    {
      directCopyDOFs_.resize(2);
      directCopyDOFs_[0] = 0;
      directCopyDOFs_[1] = 1;

      uniqueRowMatrix_[2][0] = 0.5;
      uniqueRowMatrix_[2][1] = 0.5;

      rowDOFPtrs_[0] = parentDOFs_ + 0;
      rowDOFPtrs_[1] = parentDOFs_ + 1;
      rowDOFPtrs_[2] = child0DOFs_ + 1;

      colDOFPtrs_[0] = parentDOFs_ + 0;
      colDOFPtrs_[1] = parentDOFs_ + 1;
    }
  };


  /** \brief
   * Defines the interpolation and restriction matrix for 2d-elements
   * and 1d-lagrange basis functions.
   */
  class IM_2d_1 : public InterpolRestrictMatrix
  {
  public:
    IM_2d_1() : 
      InterpolRestrictMatrix(3, 4, 3, false) 
    {}

    void fillMemory()
    {
      directCopyDOFs_.resize(3);
      directCopyDOFs_[0] = 0;
      directCopyDOFs_[1] = 1;
      directCopyDOFs_[2] = 2;

      uniqueRowMatrix_[3][0] = 0.5;
      uniqueRowMatrix_[3][1] = 0.5;

      rowDOFPtrs_[0] = parentDOFs_ + 0;
      rowDOFPtrs_[1] = parentDOFs_ + 1;
      rowDOFPtrs_[2] = parentDOFs_ + 2;
      rowDOFPtrs_[3] = child0DOFs_ + 2;

      colDOFPtrs_[0] = parentDOFs_ + 0;
      colDOFPtrs_[1] = parentDOFs_ + 1;
      colDOFPtrs_[2] = parentDOFs_ + 2;
    }
  };


  /** \brief
   * Defines the interpolation and restriction matrix for 2d-elements
   * and 2d-lagrange basis functions.
   */
  class IM_2d_2 : public InterpolRestrictMatrix
  {
  public:
    IM_2d_2() : 
      InterpolRestrictMatrix(6, 9, 6, false) 
    {}

    void fillMemory()
    {
      directCopyDOFs_.push_back(0);
      directCopyDOFs_.push_back(1);
      directCopyDOFs_.push_back(2);
      directCopyDOFs_.push_back(3);
      directCopyDOFs_.push_back(4);

      uniqueRowMatrix_[5][5] = 1.0;

      uniqueRowMatrix_[6][0] = 0.375;
      uniqueRowMatrix_[6][1] = -0.125;
      uniqueRowMatrix_[6][5] = 0.75;

      uniqueRowMatrix_[7][0] = -0.125;
      uniqueRowMatrix_[7][1] = -0.125;
      uniqueRowMatrix_[7][3] = 0.5;
      uniqueRowMatrix_[7][4] = 0.5;
      uniqueRowMatrix_[7][5] = 0.25;

      uniqueRowMatrix_[8][0] = -0.125;
      uniqueRowMatrix_[8][1] = 0.375;
      uniqueRowMatrix_[8][5] = 0.75;

      rowDOFPtrs_[0] = parentDOFs_ + 0;
      rowDOFPtrs_[1] = parentDOFs_ + 1;
      rowDOFPtrs_[2] = parentDOFs_ + 2;
      rowDOFPtrs_[3] = parentDOFs_ + 3;
      rowDOFPtrs_[4] = parentDOFs_ + 4;
      rowDOFPtrs_[5] = child0DOFs_ + 2;
      rowDOFPtrs_[6] = child0DOFs_ + 3;
      rowDOFPtrs_[7] = child0DOFs_ + 4;
      rowDOFPtrs_[8] = child1DOFs_ + 4;

      colDOFPtrs_[0] = parentDOFs_ + 0;
      colDOFPtrs_[1] = parentDOFs_ + 1;
      colDOFPtrs_[2] = parentDOFs_ + 2;
      colDOFPtrs_[3] = parentDOFs_ + 3;
      colDOFPtrs_[4] = parentDOFs_ + 4;
      colDOFPtrs_[5] = parentDOFs_ + 5;
    }
  };


  /** \brief
   * Defines the interpolation and restriction matrix for 2d-elements
   * and 3d-lagrange basis functions.
   */
  class IM_2d_3 : public InterpolRestrictMatrix
  {
  public:
    IM_2d_3() : 
      InterpolRestrictMatrix(10, 16, 10, false) 
    {}

    void fillMemory()
    {
      directCopyDOFs_.push_back(0);
      directCopyDOFs_.push_back(1);
      directCopyDOFs_.push_back(2);
      directCopyDOFs_.push_back(3);
      directCopyDOFs_.push_back(4);
      directCopyDOFs_.push_back(5);
      directCopyDOFs_.push_back(6);

      uniqueRowMatrix_[7][7] = 1.0;
      uniqueRowMatrix_[8][8] = 1.0;
      uniqueRowMatrix_[9][9] = 1.0;

      uniqueRowMatrix_[10][0] = -0.0625;
      uniqueRowMatrix_[10][1] = -0.0625;
      uniqueRowMatrix_[10][7] = 0.5625;
      uniqueRowMatrix_[10][8] = 0.5625;

      uniqueRowMatrix_[11][0] = 0.3125;
      uniqueRowMatrix_[11][1] = 0.0625;
      uniqueRowMatrix_[11][7] = 0.9375;
      uniqueRowMatrix_[11][8] = -0.3125;

      uniqueRowMatrix_[12][0] = 0.0625;
      uniqueRowMatrix_[12][1] = 0.0625;
      uniqueRowMatrix_[12][3] = -0.25;
      uniqueRowMatrix_[12][4] = 0.5;
      uniqueRowMatrix_[12][5] = 0.5;
      uniqueRowMatrix_[12][6] = -0.25;
      uniqueRowMatrix_[12][7] = -0.0625;
      uniqueRowMatrix_[12][8] = -0.0625;
      uniqueRowMatrix_[12][9] = 0.5;

      uniqueRowMatrix_[13][0] = -0.0625;
      uniqueRowMatrix_[13][1] = 0.0625;
      uniqueRowMatrix_[13][3] = -0.125;
      uniqueRowMatrix_[13][6] = 0.375;
      uniqueRowMatrix_[13][7] = 0.1875;
      uniqueRowMatrix_[13][8] = -0.1875;
      uniqueRowMatrix_[13][9] = 0.75;

      uniqueRowMatrix_[14][0] = 0.0625;
      uniqueRowMatrix_[14][1] = 0.3125;
      uniqueRowMatrix_[14][7] = -0.3125;
      uniqueRowMatrix_[14][8] = 0.9375;

      uniqueRowMatrix_[15][0] = 0.0625;
      uniqueRowMatrix_[15][1] = -0.0625;
      uniqueRowMatrix_[15][3] = 0.375;
      uniqueRowMatrix_[15][6] = -0.125;
      uniqueRowMatrix_[15][7] = -0.1875;
      uniqueRowMatrix_[15][8] = 0.1875;
      uniqueRowMatrix_[15][9] = 0.75;

      rowDOFPtrs_[ 0] = parentDOFs_ + 0;
      rowDOFPtrs_[ 1] = parentDOFs_ + 1;
      rowDOFPtrs_[ 2] = parentDOFs_ + 2;
      rowDOFPtrs_[ 3] = parentDOFs_ + 3;
      rowDOFPtrs_[ 4] = parentDOFs_ + 4;
      rowDOFPtrs_[ 5] = parentDOFs_ + 5;
      rowDOFPtrs_[ 6] = parentDOFs_ + 6;
      rowDOFPtrs_[ 7] = child0DOFs_ + 4;
      rowDOFPtrs_[ 8] = child1DOFs_ + 5;
      rowDOFPtrs_[ 9] = child0DOFs_ + 5;
      rowDOFPtrs_[10] = child0DOFs_ + 2;
      rowDOFPtrs_[11] = child0DOFs_ + 3;
      rowDOFPtrs_[12] = child0DOFs_ + 6;
      rowDOFPtrs_[13] = child0DOFs_ + 9;
      rowDOFPtrs_[14] = child1DOFs_ + 6;
      rowDOFPtrs_[15] = child1DOFs_ + 9;

      colDOFPtrs_[0] = parentDOFs_ + 0;
      colDOFPtrs_[1] = parentDOFs_ + 1;
      colDOFPtrs_[2] = parentDOFs_ + 2;
      colDOFPtrs_[3] = parentDOFs_ + 3;
      colDOFPtrs_[4] = parentDOFs_ + 4;
      colDOFPtrs_[5] = parentDOFs_ + 5;
      colDOFPtrs_[6] = parentDOFs_ + 6;
      colDOFPtrs_[7] = parentDOFs_ + 7;
      colDOFPtrs_[8] = parentDOFs_ + 8;
      colDOFPtrs_[9] = parentDOFs_ + 9;
    }
  };


  /** \brief
   * Defines the interpolation and restriction matrix for 2d-elements
   * and 4d-lagrange basis functions.
   */
  class IM_2d_4 : public InterpolRestrictMatrix
  {
  public:
    IM_2d_4() : 
      InterpolRestrictMatrix(15, 25, 15, false) 
    {}

    void fillMemory()
    {
      directCopyDOFs_.push_back(0);
      directCopyDOFs_.push_back(1);
      directCopyDOFs_.push_back(2);
      directCopyDOFs_.push_back(3);
      directCopyDOFs_.push_back(4);
      directCopyDOFs_.push_back(5);
      directCopyDOFs_.push_back(6);
      directCopyDOFs_.push_back(7);
      directCopyDOFs_.push_back(8);

      uniqueRowMatrix_[ 9][ 9] = 1.0;
      uniqueRowMatrix_[10][10] = 1.0;
      uniqueRowMatrix_[11][11] = 1.0;
      uniqueRowMatrix_[12][12] = 1.0;
      uniqueRowMatrix_[13][13] = 1.0;
      uniqueRowMatrix_[14][14] = 1.0;

      uniqueRowMatrix_[15][ 0] = 0.2734375;
      uniqueRowMatrix_[15][ 1] = -0.0390625;
      uniqueRowMatrix_[15][ 9] = 1.09375;
      uniqueRowMatrix_[15][10] = -0.546875;
      uniqueRowMatrix_[15][11] = 0.21875;

      uniqueRowMatrix_[16][ 0] = -0.0390625;
      uniqueRowMatrix_[16][ 1] = 0.0234375;
      uniqueRowMatrix_[16][ 9] = 0.46875;
      uniqueRowMatrix_[16][10] = 0.703125;
      uniqueRowMatrix_[16][11] = -0.15625;

      uniqueRowMatrix_[17][ 0] = 0.0234375;
      uniqueRowMatrix_[17][ 1] = 0.0234375;
      uniqueRowMatrix_[17][ 3] = -0.0625;
      uniqueRowMatrix_[17][ 8] = -0.0625;
      uniqueRowMatrix_[17][ 9] = -0.09375;
      uniqueRowMatrix_[17][10] = 0.140625;
      uniqueRowMatrix_[17][11] = -0.09375;
      uniqueRowMatrix_[17][12] = 0.5625;
      uniqueRowMatrix_[17][13] = 0.5625;

      uniqueRowMatrix_[18][ 0] = -0.0390625;
      uniqueRowMatrix_[18][ 1] = -0.0390625;
      uniqueRowMatrix_[18][ 3] = 0.1875;
      uniqueRowMatrix_[18][ 4] = -0.375;
      uniqueRowMatrix_[18][ 5] = 0.5;
      uniqueRowMatrix_[18][ 6] = 0.5;
      uniqueRowMatrix_[18][ 7] = -0.375;
      uniqueRowMatrix_[18][ 8] = 0.1875;
      uniqueRowMatrix_[18][ 9] = 0.03125;
      uniqueRowMatrix_[18][10] = 0.015625;
      uniqueRowMatrix_[18][11] = 0.03125;
      uniqueRowMatrix_[18][12] = -0.1875;
      uniqueRowMatrix_[18][13] = -0.1875;
      uniqueRowMatrix_[18][14] = 0.75;

      uniqueRowMatrix_[19][ 0] = 0.0234375;
      uniqueRowMatrix_[19][ 1] = -0.0390625;
      uniqueRowMatrix_[19][ 3] = 0.125;
      uniqueRowMatrix_[19][ 4] = -0.125;
      uniqueRowMatrix_[19][ 7] = 0.375;
      uniqueRowMatrix_[19][ 8] = -0.125;
      uniqueRowMatrix_[19][ 9] = -0.03125;
      uniqueRowMatrix_[19][10] = -0.046875;
      uniqueRowMatrix_[19][11] = 0.09375;
      uniqueRowMatrix_[19][12] = 0.375;
      uniqueRowMatrix_[19][13] = -0.375;
      uniqueRowMatrix_[19][14] = 0.75;

      uniqueRowMatrix_[20][ 0] = -0.0390625;
      uniqueRowMatrix_[20][ 1] = -0.0390625;
      uniqueRowMatrix_[20][ 3] = 0.0625;
      uniqueRowMatrix_[20][ 8] = 0.3125;
      uniqueRowMatrix_[20][13] = -0.3125;
      uniqueRowMatrix_[20][ 9] = 0.15625;
      uniqueRowMatrix_[20][11] = 0.15625;
      uniqueRowMatrix_[20][10] = -0.234375;
      uniqueRowMatrix_[20][12] = 0.9375;

      uniqueRowMatrix_[21][ 0] = 0.0234375;
      uniqueRowMatrix_[21][ 1] = -0.0390625;
      uniqueRowMatrix_[21][ 9] = -0.15625;
      uniqueRowMatrix_[21][10] = 0.703125;
      uniqueRowMatrix_[21][11] = 0.46875;

      uniqueRowMatrix_[22][ 0] = -0.0390625;
      uniqueRowMatrix_[22][ 1] = 0.2734375;
      uniqueRowMatrix_[22][ 9] = 0.21875;
      uniqueRowMatrix_[22][10] = -0.546875;
      uniqueRowMatrix_[22][11] = 1.09375;

      uniqueRowMatrix_[23][ 0] = -0.0390625;
      uniqueRowMatrix_[23][ 1] = -0.0390625;
      uniqueRowMatrix_[23][ 3] = 0.3125;
      uniqueRowMatrix_[23][ 8] = 0.0625;
      uniqueRowMatrix_[23][ 9] = 0.15625;
      uniqueRowMatrix_[23][10] = -0.234375;
      uniqueRowMatrix_[23][11] = 0.15625;
      uniqueRowMatrix_[23][12] = -0.3125;
      uniqueRowMatrix_[23][13] = 0.9375;

      uniqueRowMatrix_[24][ 0] = -0.0390625;
      uniqueRowMatrix_[24][ 1] = 0.0234375;
      uniqueRowMatrix_[24][ 3] = -0.125;
      uniqueRowMatrix_[24][ 7] = -0.125;
      uniqueRowMatrix_[24][ 8] = 0.125;
      uniqueRowMatrix_[24][ 4] = 0.375;
      uniqueRowMatrix_[24][12] = -0.375;
      uniqueRowMatrix_[24][13] = 0.375;
      uniqueRowMatrix_[24][ 9] = 0.09375;
      uniqueRowMatrix_[24][10] = -0.046875;
      uniqueRowMatrix_[24][11] = -0.03125;
      uniqueRowMatrix_[24][14] = 0.75;

      rowDOFPtrs_[ 0] = parentDOFs_ + 0;
      rowDOFPtrs_[ 1] = parentDOFs_ + 1;
      rowDOFPtrs_[ 2] = parentDOFs_ + 2;
      rowDOFPtrs_[ 3] = parentDOFs_ + 3;
      rowDOFPtrs_[ 4] = parentDOFs_ + 4;
      rowDOFPtrs_[ 5] = parentDOFs_ + 5;
      rowDOFPtrs_[ 6] = parentDOFs_ + 6;
      rowDOFPtrs_[ 7] = parentDOFs_ + 7;
      rowDOFPtrs_[ 8] = parentDOFs_ + 8;
      rowDOFPtrs_[ 9] = child0DOFs_ + 4;
      rowDOFPtrs_[10] = child0DOFs_ + 2;
      rowDOFPtrs_[11] = child1DOFs_ + 7;
      rowDOFPtrs_[12] = child0DOFs_ + 14;
      rowDOFPtrs_[13] = child1DOFs_ + 14;
      rowDOFPtrs_[14] = child0DOFs_ + 7;
      rowDOFPtrs_[15] = child0DOFs_ + 3;
      rowDOFPtrs_[16] = child0DOFs_ + 5;
      rowDOFPtrs_[17] = child0DOFs_ + 6;
      rowDOFPtrs_[18] = child0DOFs_ + 8;
      rowDOFPtrs_[19] = child0DOFs_ + 12;
      rowDOFPtrs_[20] = child0DOFs_ + 13;
      rowDOFPtrs_[21] = child1DOFs_ + 6;
      rowDOFPtrs_[22] = child1DOFs_ + 8;
      rowDOFPtrs_[23] = child1DOFs_ + 12;
      rowDOFPtrs_[24] = child1DOFs_ + 13;

      colDOFPtrs_[ 0] = parentDOFs_ +  0;
      colDOFPtrs_[ 1] = parentDOFs_ +  1;
      colDOFPtrs_[ 2] = parentDOFs_ +  2;
      colDOFPtrs_[ 3] = parentDOFs_ +  3;
      colDOFPtrs_[ 4] = parentDOFs_ +  4;
      colDOFPtrs_[ 5] = parentDOFs_ +  5;
      colDOFPtrs_[ 6] = parentDOFs_ +  6;
      colDOFPtrs_[ 7] = parentDOFs_ +  7;
      colDOFPtrs_[ 8] = parentDOFs_ +  8;
      colDOFPtrs_[ 9] = parentDOFs_ +  9;
      colDOFPtrs_[10] = parentDOFs_ + 10;
      colDOFPtrs_[11] = parentDOFs_ + 11;
      colDOFPtrs_[12] = parentDOFs_ + 12;
      colDOFPtrs_[13] = parentDOFs_ + 13;
      colDOFPtrs_[14] = parentDOFs_ + 14;
    }
  };



  /** \brief
   * Defines the interpolation and restriction matrix for 3d-elements
   * and 1d-lagrange basis functions.
   */
  class IM_3d_1 : public InterpolRestrictMatrix
  {
  public:
    IM_3d_1() : 
      InterpolRestrictMatrix(4, 5, 4, false) 
    {}

    void fillMemory()
    {
      directCopyDOFs_.resize(4);
      directCopyDOFs_[0] = 0;
      directCopyDOFs_[1] = 1;
      directCopyDOFs_[2] = 2;
      directCopyDOFs_[3] = 3;

      uniqueRowMatrix_[4][0] = 0.5;
      uniqueRowMatrix_[4][1] = 0.5;

      rowDOFPtrs_[0] = parentDOFs_ + 0;
      rowDOFPtrs_[1] = parentDOFs_ + 1;
      rowDOFPtrs_[2] = parentDOFs_ + 2;
      rowDOFPtrs_[3] = parentDOFs_ + 3;
      rowDOFPtrs_[4] = child0DOFs_ + 3;

      colDOFPtrs_[0] = parentDOFs_ + 0;
      colDOFPtrs_[1] = parentDOFs_ + 1;
      colDOFPtrs_[2] = parentDOFs_ + 2;
      colDOFPtrs_[3] = parentDOFs_ + 3;
    }
  };



  /** \brief
   * Defines the interpolation and restriction matrix for 3d-elements
   * and 2d-lagrange basis functions.
   */
  class IM_3d_2 : public InterpolRestrictMatrix
  {
  public:
    IM_3d_2() : 
      InterpolRestrictMatrix(10, 14, 10, false) 
    {}

    void fillMemory()
    {
      directCopyDOFs_.push_back(0);
      directCopyDOFs_.push_back(1);
      directCopyDOFs_.push_back(2);
      directCopyDOFs_.push_back(3);
      directCopyDOFs_.push_back(5);
      directCopyDOFs_.push_back(6);
      directCopyDOFs_.push_back(7);
      directCopyDOFs_.push_back(8);
      directCopyDOFs_.push_back(9);

      uniqueRowMatrix_[4][4] = 1.0;

      uniqueRowMatrix_[10][0] = 0.375;
      uniqueRowMatrix_[10][1] = -0.125;
      uniqueRowMatrix_[10][4] = 0.75;

      uniqueRowMatrix_[11][0] = -0.125;
      uniqueRowMatrix_[11][1] = -0.125;
      uniqueRowMatrix_[11][4] = 0.25;
      uniqueRowMatrix_[11][5] = 0.5;
      uniqueRowMatrix_[11][7] = 0.5;

      uniqueRowMatrix_[12][0] = -0.125;
      uniqueRowMatrix_[12][1] = -0.125;
      uniqueRowMatrix_[12][4] = 0.25;
      uniqueRowMatrix_[12][6] = 0.5;
      uniqueRowMatrix_[12][8] = 0.5;

      uniqueRowMatrix_[13][0] = -0.125;
      uniqueRowMatrix_[13][1] = 0.375;
      uniqueRowMatrix_[13][4] = 0.75;


      rowDOFPtrs_[0] = parentDOFs_ + 0;
      rowDOFPtrs_[1] = parentDOFs_ + 1;
      rowDOFPtrs_[2] = parentDOFs_ + 2;
      rowDOFPtrs_[3] = parentDOFs_ + 3;
      rowDOFPtrs_[4] = child0DOFs_ + 3;
      rowDOFPtrs_[5] = parentDOFs_ + 5;
      rowDOFPtrs_[6] = parentDOFs_ + 6;
      rowDOFPtrs_[7] = parentDOFs_ + 7;
      rowDOFPtrs_[8] = parentDOFs_ + 8;
      rowDOFPtrs_[9] = parentDOFs_ + 9;
      rowDOFPtrs_[10] = child0DOFs_ + 6;
      rowDOFPtrs_[11] = child0DOFs_ + 8;
      rowDOFPtrs_[12] = child0DOFs_ + 9;
      rowDOFPtrs_[13] = child1DOFs_ + 6;

      colDOFPtrs_[0] = parentDOFs_ + 0;
      colDOFPtrs_[1] = parentDOFs_ + 1;
      colDOFPtrs_[2] = parentDOFs_ + 2;
      colDOFPtrs_[3] = parentDOFs_ + 3;
      colDOFPtrs_[4] = parentDOFs_ + 4;
      colDOFPtrs_[5] = parentDOFs_ + 5;
      colDOFPtrs_[6] = parentDOFs_ + 6;
      colDOFPtrs_[7] = parentDOFs_ + 7;
      colDOFPtrs_[8] = parentDOFs_ + 8;
      colDOFPtrs_[9] = parentDOFs_ + 9;
    }
  };



  /** \brief
   * Defines the interpolation and restriction matrix for 3d-elements
   * and 3d-lagrange basis functions.
   */
  class IM_3d_3 : public InterpolRestrictMatrix
  {
  public:
    IM_3d_3() : 
      InterpolRestrictMatrix(20, 30, 20, true) 
    {}

    void fillMemory()
    {
      directCopyDOFs_.push_back(0);
      directCopyDOFs_.push_back(1);
      directCopyDOFs_.push_back(2);
      directCopyDOFs_.push_back(3);
      directCopyDOFs_.push_back(6);
      directCopyDOFs_.push_back(7);
      directCopyDOFs_.push_back(8);
      directCopyDOFs_.push_back(9);
      directCopyDOFs_.push_back(10);
      directCopyDOFs_.push_back(11);
      directCopyDOFs_.push_back(12);
      directCopyDOFs_.push_back(13);
      directCopyDOFs_.push_back(14);
      directCopyDOFs_.push_back(15);
      directCopyDOFs_.push_back(16);
      directCopyDOFs_.push_back(17);

      uniqueRowMatrix_[ 4][ 4] = 1.0;
      uniqueRowMatrix_[ 5][ 5] = 1.0;
      uniqueRowMatrix_[18][18] = 1.0;
      uniqueRowMatrix_[19][19] = 1.0;

      uniqueRowMatrix_[20][ 0] = -0.0625;
      uniqueRowMatrix_[20][ 1] = -0.0625;
      uniqueRowMatrix_[20][ 4] = 0.5625;
      uniqueRowMatrix_[20][ 5] = 0.5625;

      uniqueRowMatrix_[21][ 0] = 0.3125;
      uniqueRowMatrix_[21][ 1] = 0.0625;
      uniqueRowMatrix_[21][ 4] = 0.9375;
      uniqueRowMatrix_[21][ 5] = -0.3125;

      uniqueRowMatrix_[22][ 0] = 0.0625;
      uniqueRowMatrix_[22][ 1] = 0.0625;
      uniqueRowMatrix_[22][ 4] = -0.0625;
      uniqueRowMatrix_[22][ 5] = -0.0625;
      uniqueRowMatrix_[22][ 6] = -0.25;
      uniqueRowMatrix_[22][ 7] = 0.5;
      uniqueRowMatrix_[22][10] = -0.25;
      uniqueRowMatrix_[22][11] = 0.5;
      uniqueRowMatrix_[22][19] = 0.5;

      uniqueRowMatrix_[23][ 0] = 0.0625;
      uniqueRowMatrix_[23][ 1] = 0.0625;
      uniqueRowMatrix_[23][ 4] = -0.0625;
      uniqueRowMatrix_[23][ 5] = -0.0625;
      uniqueRowMatrix_[23][ 8] = -0.25;
      uniqueRowMatrix_[23][ 9] = 0.5;
      uniqueRowMatrix_[23][12] = -0.25;
      uniqueRowMatrix_[23][13] = 0.5;
      uniqueRowMatrix_[23][18] = 0.5;

      uniqueRowMatrix_[24][ 0] = 0.0625;
      uniqueRowMatrix_[24][ 1] = 0.0625;
      uniqueRowMatrix_[24][ 4] = -0.0625;
      uniqueRowMatrix_[24][ 5] = -0.0625;
      uniqueRowMatrix_[24][ 6] = -0.125;
      uniqueRowMatrix_[24][ 8] = -0.125;
      uniqueRowMatrix_[24][10] = -0.125;
      uniqueRowMatrix_[24][12] = -0.125;
      uniqueRowMatrix_[24][16] = 0.5;
      uniqueRowMatrix_[24][17] = 0.5;
      uniqueRowMatrix_[24][18] = 0.25;
      uniqueRowMatrix_[24][19] = 0.25;

      uniqueRowMatrix_[25][ 0] = -0.0625;
      uniqueRowMatrix_[25][ 1] = 0.0625;
      uniqueRowMatrix_[25][ 4] = 0.1875;
      uniqueRowMatrix_[25][ 5] = -0.1875;
      uniqueRowMatrix_[25][ 8] = 0.375;
      uniqueRowMatrix_[25][12] = -0.125;
      uniqueRowMatrix_[25][18] = 0.75;

      uniqueRowMatrix_[26][ 0] = -0.0625;
      uniqueRowMatrix_[26][ 1] = 0.0625;
      uniqueRowMatrix_[26][ 4] = 0.1875;
      uniqueRowMatrix_[26][ 5] = -0.1875;
      uniqueRowMatrix_[26][ 6] = 0.375;
      uniqueRowMatrix_[26][10] = -0.125;
      uniqueRowMatrix_[26][19] = 0.75;

      uniqueRowMatrix_[27][ 0] = 0.0625;
      uniqueRowMatrix_[27][ 1] = 0.3125;
      uniqueRowMatrix_[27][ 4] = -0.3125;
      uniqueRowMatrix_[27][ 5] = 0.9375;

      uniqueRowMatrix_[28][ 0] = 0.0625;
      uniqueRowMatrix_[28][ 1] = -0.0625;
      uniqueRowMatrix_[28][ 4] = -0.1875;
      uniqueRowMatrix_[28][ 5] = 0.1875;
      uniqueRowMatrix_[28][ 8] = -0.125;
      uniqueRowMatrix_[28][12] = 0.375;
      uniqueRowMatrix_[28][18] = 0.75;

      uniqueRowMatrix_[29][ 0] = 0.0625;
      uniqueRowMatrix_[29][ 1] = -0.0625;
      uniqueRowMatrix_[29][ 4] = -0.1875;
      uniqueRowMatrix_[29][ 5] = 0.1875;
      uniqueRowMatrix_[29][ 6] = -0.125;
      uniqueRowMatrix_[29][10] = 0.375;
      uniqueRowMatrix_[29][19] = 0.75;

      rowDOFPtrs_[0] = parentDOFs_ + 0;
      rowDOFPtrs_[1] = parentDOFs_ + 1;
      rowDOFPtrs_[2] = parentDOFs_ + 2;
      rowDOFPtrs_[3] = parentDOFs_ + 3;
      rowDOFPtrs_[4] = child0DOFs_ + 9;
      rowDOFPtrs_[5] = child1DOFs_ + 9;
      rowDOFPtrs_[6] = parentDOFs_ + 6;
      rowDOFPtrs_[7] = parentDOFs_ + 7;
      rowDOFPtrs_[8] = parentDOFs_ + 8;
      rowDOFPtrs_[9] = parentDOFs_ + 9;
      rowDOFPtrs_[10] = parentDOFs_ + 10;
      rowDOFPtrs_[11] = parentDOFs_ + 11;
      rowDOFPtrs_[12] = parentDOFs_ + 12;
      rowDOFPtrs_[13] = parentDOFs_ + 13;
      rowDOFPtrs_[14] = parentDOFs_ + 14;
      rowDOFPtrs_[15] = parentDOFs_ + 15;
      rowDOFPtrs_[16] = parentDOFs_ + 16;
      rowDOFPtrs_[17] = parentDOFs_ + 17;
      rowDOFPtrs_[18] = child0DOFs_ + 15;
      rowDOFPtrs_[19] = child0DOFs_ + 13;
      rowDOFPtrs_[20] = child0DOFs_ + 3;
      rowDOFPtrs_[21] = child0DOFs_ + 8;
      rowDOFPtrs_[22] = child0DOFs_ + 12;
      rowDOFPtrs_[23] = child0DOFs_ + 14;
      rowDOFPtrs_[24] = child0DOFs_ + 16;
      rowDOFPtrs_[25] = child0DOFs_ + 17;
      rowDOFPtrs_[26] = child0DOFs_ + 18;
      rowDOFPtrs_[27] = child1DOFs_ + 8;
      rowDOFPtrs_[28] = child1DOFs_ + 17;
      rowDOFPtrs_[29] = child1DOFs_ + 18;

      colDOFPtrs_[ 0] = parentDOFs_ + 0;
      colDOFPtrs_[ 1] = parentDOFs_ + 1;
      colDOFPtrs_[ 2] = parentDOFs_ + 2;
      colDOFPtrs_[ 3] = parentDOFs_ + 3;
      colDOFPtrs_[ 4] = parentDOFs_ + 4;
      colDOFPtrs_[ 5] = parentDOFs_ + 5;
      colDOFPtrs_[ 6] = parentDOFs_ + 6;
      colDOFPtrs_[ 7] = parentDOFs_ + 7;
      colDOFPtrs_[ 8] = parentDOFs_ + 8;
      colDOFPtrs_[ 9] = parentDOFs_ + 9;
      colDOFPtrs_[10] = parentDOFs_ + 10;
      colDOFPtrs_[11] = parentDOFs_ + 11;
      colDOFPtrs_[12] = parentDOFs_ + 12;
      colDOFPtrs_[13] = parentDOFs_ + 13;
      colDOFPtrs_[14] = parentDOFs_ + 14;
      colDOFPtrs_[15] = parentDOFs_ + 15;
      colDOFPtrs_[16] = parentDOFs_ + 16;
      colDOFPtrs_[17] = parentDOFs_ + 17;
      colDOFPtrs_[18] = parentDOFs_ + 18;
      colDOFPtrs_[19] = parentDOFs_ + 19;

      type0_rowDOFPtrs_[0] = parentDOFs_ + 0;
      type0_rowDOFPtrs_[1] = parentDOFs_ + 1;
      type0_rowDOFPtrs_[2] = parentDOFs_ + 2;
      type0_rowDOFPtrs_[3] = parentDOFs_ + 3;
      type0_rowDOFPtrs_[4] = child0DOFs_ + 9;
      type0_rowDOFPtrs_[5] = child1DOFs_ + 9;
      type0_rowDOFPtrs_[6] = parentDOFs_ + 6;
      type0_rowDOFPtrs_[7] = parentDOFs_ + 7;
      type0_rowDOFPtrs_[8] = parentDOFs_ + 8;
      type0_rowDOFPtrs_[9] = parentDOFs_ + 9;
      type0_rowDOFPtrs_[10] = parentDOFs_ + 10;
      type0_rowDOFPtrs_[11] = parentDOFs_ + 11;
      type0_rowDOFPtrs_[12] = parentDOFs_ + 12;
      type0_rowDOFPtrs_[13] = parentDOFs_ + 13;
      type0_rowDOFPtrs_[14] = parentDOFs_ + 14;
      type0_rowDOFPtrs_[15] = parentDOFs_ + 15;
      type0_rowDOFPtrs_[16] = parentDOFs_ + 16;
      type0_rowDOFPtrs_[17] = parentDOFs_ + 17;
      type0_rowDOFPtrs_[18] = child0DOFs_ + 15;
      type0_rowDOFPtrs_[19] = child0DOFs_ + 13;
      type0_rowDOFPtrs_[20] = child0DOFs_ + 3;
      type0_rowDOFPtrs_[21] = child0DOFs_ + 8;
      type0_rowDOFPtrs_[22] = child0DOFs_ + 12;
      type0_rowDOFPtrs_[23] = child0DOFs_ + 14;
      type0_rowDOFPtrs_[24] = child0DOFs_ + 16;
      type0_rowDOFPtrs_[25] = child0DOFs_ + 17;
      type0_rowDOFPtrs_[26] = child0DOFs_ + 18;
      type0_rowDOFPtrs_[27] = child1DOFs_ + 8;
      type0_rowDOFPtrs_[28] = child1DOFs_ + 18;
      type0_rowDOFPtrs_[29] = child1DOFs_ + 17;

      type0_colDOFPtrs_[ 0] = parentDOFs_ + 0;
      type0_colDOFPtrs_[ 1] = parentDOFs_ + 1;
      type0_colDOFPtrs_[ 2] = parentDOFs_ + 2;
      type0_colDOFPtrs_[ 3] = parentDOFs_ + 3;
      type0_colDOFPtrs_[ 4] = parentDOFs_ + 4;
      type0_colDOFPtrs_[ 5] = parentDOFs_ + 5;
      type0_colDOFPtrs_[ 6] = parentDOFs_ + 6;
      type0_colDOFPtrs_[ 7] = parentDOFs_ + 7;
      type0_colDOFPtrs_[ 8] = parentDOFs_ + 8;
      type0_colDOFPtrs_[ 9] = parentDOFs_ + 9;
      type0_colDOFPtrs_[10] = parentDOFs_ + 10;
      type0_colDOFPtrs_[11] = parentDOFs_ + 11;
      type0_colDOFPtrs_[12] = parentDOFs_ + 12;
      type0_colDOFPtrs_[13] = parentDOFs_ + 13;
      type0_colDOFPtrs_[14] = parentDOFs_ + 14;
      type0_colDOFPtrs_[15] = parentDOFs_ + 15;
      type0_colDOFPtrs_[16] = parentDOFs_ + 16;
      type0_colDOFPtrs_[17] = parentDOFs_ + 17;
      type0_colDOFPtrs_[18] = parentDOFs_ + 18;
      type0_colDOFPtrs_[19] = parentDOFs_ + 19;
    }
  };


  /** \brief
   * Defines the interpolation and restriction matrix for 3d-elements
   * and 4d-lagrange basis functions.
   */
  class IM_3d_4 : public InterpolRestrictMatrix
  {
  public:
    IM_3d_4() : 
      InterpolRestrictMatrix(35, 55, 35, true) 
    {}

    void fillMemory()
    {
      directCopyDOFs_.push_back(0);
      directCopyDOFs_.push_back(1);
      directCopyDOFs_.push_back(2);
      directCopyDOFs_.push_back(3);
      directCopyDOFs_.push_back(7);
      directCopyDOFs_.push_back(8);
      directCopyDOFs_.push_back(9);
      directCopyDOFs_.push_back(10);
      directCopyDOFs_.push_back(11);
      directCopyDOFs_.push_back(12);
      directCopyDOFs_.push_back(13);
      directCopyDOFs_.push_back(14);
      directCopyDOFs_.push_back(15);
      directCopyDOFs_.push_back(16);
      directCopyDOFs_.push_back(17);
      directCopyDOFs_.push_back(18);
      directCopyDOFs_.push_back(19);
      directCopyDOFs_.push_back(20);
      directCopyDOFs_.push_back(21);
      directCopyDOFs_.push_back(22);
      directCopyDOFs_.push_back(23);
      directCopyDOFs_.push_back(24);
      directCopyDOFs_.push_back(25);
      directCopyDOFs_.push_back(26);
      directCopyDOFs_.push_back(27);

      uniqueRowMatrix_[ 4][ 4] = 1.0;
      uniqueRowMatrix_[ 5][ 5] = 1.0;
      uniqueRowMatrix_[ 6][ 6] = 1.0;
      uniqueRowMatrix_[28][28] = 1.0;
      uniqueRowMatrix_[29][29] = 1.0;
      uniqueRowMatrix_[30][30] = 1.0;
      uniqueRowMatrix_[31][31] = 1.0;
      uniqueRowMatrix_[32][32] = 1.0;
      uniqueRowMatrix_[33][33] = 1.0;
      uniqueRowMatrix_[34][34] = 1.0;

      uniqueRowMatrix_[35][ 0] = 0.2734375;
      uniqueRowMatrix_[35][ 1] = -0.0390625;
      uniqueRowMatrix_[35][ 4] = 1.09375;
      uniqueRowMatrix_[35][ 5] = -0.546875;
      uniqueRowMatrix_[35][ 6] = 0.21875;

      uniqueRowMatrix_[36][ 0] = -0.0390625;
      uniqueRowMatrix_[36][ 1] = 0.0234375;
      uniqueRowMatrix_[36][ 4] = 0.46875;
      uniqueRowMatrix_[36][ 5] = 0.703125;
      uniqueRowMatrix_[36][ 6] = -0.15625;

      uniqueRowMatrix_[37][ 0] = -0.0390625;
      uniqueRowMatrix_[37][ 1] = -0.0390625;
      uniqueRowMatrix_[37][ 4] = 0.03125;
      uniqueRowMatrix_[37][ 5] = 0.015625;
      uniqueRowMatrix_[37][ 6] = 0.03125;
      uniqueRowMatrix_[37][ 7] = 0.1875;
      uniqueRowMatrix_[37][ 8] = -0.375;
      uniqueRowMatrix_[37][ 9] = 0.5;
      uniqueRowMatrix_[37][13] = 0.1875;
      uniqueRowMatrix_[37][14] = -0.375;
      uniqueRowMatrix_[37][15] = 0.5;
      uniqueRowMatrix_[37][31] = -0.1875;
      uniqueRowMatrix_[37][32] = -0.1875;
      uniqueRowMatrix_[37][33] = 0.75;

      uniqueRowMatrix_[38][ 0] = 0.0234375;
      uniqueRowMatrix_[38][ 1] = 0.0234375;
      uniqueRowMatrix_[38][ 4] = -0.09375;
      uniqueRowMatrix_[38][ 5] = 0.140625;
      uniqueRowMatrix_[38][ 6] = -0.09375;
      uniqueRowMatrix_[38][ 7] = -0.0625;
      uniqueRowMatrix_[38][13] = -0.0625;
      uniqueRowMatrix_[38][31] = 0.5625;
      uniqueRowMatrix_[38][32] = 0.5625;

      uniqueRowMatrix_[39][ 0] = -0.0390625;
      uniqueRowMatrix_[39][ 1] = -0.0390625;
      uniqueRowMatrix_[39][ 4] = 0.03125;
      uniqueRowMatrix_[39][ 5] = 0.015625;
      uniqueRowMatrix_[39][ 6] = 0.03125;
      uniqueRowMatrix_[39][10] = 0.1875;
      uniqueRowMatrix_[39][11] = -0.375;
      uniqueRowMatrix_[39][12] = 0.5;
      uniqueRowMatrix_[39][16] = 0.1875;
      uniqueRowMatrix_[39][17] = -0.375;
      uniqueRowMatrix_[39][18] = 0.5;
      uniqueRowMatrix_[39][28] = -0.1875;
      uniqueRowMatrix_[39][29] = -0.1875;
      uniqueRowMatrix_[39][30] = 0.75;

      uniqueRowMatrix_[40][ 0] = 0.0234375;
      uniqueRowMatrix_[40][ 1] = 0.0234375;
      uniqueRowMatrix_[40][ 4] = -0.09375;
      uniqueRowMatrix_[40][ 5] = 0.140625;
      uniqueRowMatrix_[40][ 6] = -0.09375;
      uniqueRowMatrix_[40][10] = -0.0625;
      uniqueRowMatrix_[40][16] = -0.0625;
      uniqueRowMatrix_[40][28] = 0.5625;
      uniqueRowMatrix_[40][29] = 0.5625;

      uniqueRowMatrix_[41][ 0] = -0.0390625;
      uniqueRowMatrix_[41][ 1] = -0.0390625;
      uniqueRowMatrix_[41][ 4] = 0.03125;
      uniqueRowMatrix_[41][ 5] = 0.015625;
      uniqueRowMatrix_[41][ 6] = 0.03125;
      uniqueRowMatrix_[41][ 7] = 0.125;
      uniqueRowMatrix_[41][ 8] = -0.125;
      uniqueRowMatrix_[41][10] = 0.0625;
      uniqueRowMatrix_[41][13] = 0.125;
      uniqueRowMatrix_[41][14] = -0.125;
      uniqueRowMatrix_[41][16] = 0.0625;
      uniqueRowMatrix_[41][22] = -0.25;
      uniqueRowMatrix_[41][23] = 0.5;
      uniqueRowMatrix_[41][25] = -0.25;
      uniqueRowMatrix_[41][26] = 0.5;
      uniqueRowMatrix_[41][28] = -0.0625;
      uniqueRowMatrix_[41][29] = -0.0625;
      uniqueRowMatrix_[41][31] = -0.125;
      uniqueRowMatrix_[41][32] = -0.125;
      uniqueRowMatrix_[41][33] = 0.25;
      uniqueRowMatrix_[41][34] = 0.5;

      uniqueRowMatrix_[42][ 0] = -0.0390625;
      uniqueRowMatrix_[42][ 1] = -0.0390625;
      uniqueRowMatrix_[42][ 4] = 0.03125;
      uniqueRowMatrix_[42][ 5] = 0.015625;
      uniqueRowMatrix_[42][ 6] = 0.03125;
      uniqueRowMatrix_[42][ 7] = 0.0625;
      uniqueRowMatrix_[42][10] = 0.125;
      uniqueRowMatrix_[42][11] = -0.125;
      uniqueRowMatrix_[42][13] = 0.0625;
      uniqueRowMatrix_[42][16] = 0.125;
      uniqueRowMatrix_[42][17] = -0.125;
      uniqueRowMatrix_[42][22] = -0.25;
      uniqueRowMatrix_[42][24] = 0.5;
      uniqueRowMatrix_[42][25] = -0.25;
      uniqueRowMatrix_[42][27] = 0.5;
      uniqueRowMatrix_[42][28] = -0.125;
      uniqueRowMatrix_[42][29] = -0.125;
      uniqueRowMatrix_[42][30] = 0.25;
      uniqueRowMatrix_[42][31] = -0.0625;
      uniqueRowMatrix_[42][32] = -0.0625;
      uniqueRowMatrix_[42][34] = 0.5;

      uniqueRowMatrix_[43][ 0] = -0.0390625;
      uniqueRowMatrix_[43][ 1] = -0.0390625;
      uniqueRowMatrix_[43][ 4] = 0.15625;
      uniqueRowMatrix_[43][ 5] = -0.234375;
      uniqueRowMatrix_[43][ 6] = 0.15625;
      uniqueRowMatrix_[43][10] = 0.3125;
      uniqueRowMatrix_[43][16] = 0.0625;
      uniqueRowMatrix_[43][28] = 0.9375;
      uniqueRowMatrix_[43][29] = -0.3125;

      uniqueRowMatrix_[44][ 0] = 0.0234375;
      uniqueRowMatrix_[44][ 1] = -0.0390625;
      uniqueRowMatrix_[44][ 4] = -0.03125;
      uniqueRowMatrix_[44][ 5] = -0.046875;
      uniqueRowMatrix_[44][ 6] = 0.09375;
      uniqueRowMatrix_[44][10] = -0.125;
      uniqueRowMatrix_[44][11] = 0.375;
      uniqueRowMatrix_[44][16] = 0.125;
      uniqueRowMatrix_[44][17] = -0.125;
      uniqueRowMatrix_[44][28] = 0.375;
      uniqueRowMatrix_[44][29] = -0.375;
      uniqueRowMatrix_[44][30] = 0.75;

      uniqueRowMatrix_[45][ 0] = -0.0390625;
      uniqueRowMatrix_[45][ 1] = -0.0390625;
      uniqueRowMatrix_[45][ 4] = 0.15625;
      uniqueRowMatrix_[45][ 5] = -0.234375;
      uniqueRowMatrix_[45][ 6] = 0.15625;
      uniqueRowMatrix_[45][ 7] = 0.3125;
      uniqueRowMatrix_[45][13] = 0.0625;
      uniqueRowMatrix_[45][31] = 0.9375;
      uniqueRowMatrix_[45][32] = -0.3125;

      uniqueRowMatrix_[46][ 0] = 0.0234375;
      uniqueRowMatrix_[46][ 1] = -0.0390625;
      uniqueRowMatrix_[46][ 4] = -0.03125;
      uniqueRowMatrix_[46][ 5] = -0.046875;
      uniqueRowMatrix_[46][ 6] = 0.09375;
      uniqueRowMatrix_[46][ 7] = -0.125;
      uniqueRowMatrix_[46][ 8] = 0.375;
      uniqueRowMatrix_[46][13] = 0.125;
      uniqueRowMatrix_[46][14] = -0.125;
      uniqueRowMatrix_[46][31] = 0.375;
      uniqueRowMatrix_[46][32] = -0.375;
      uniqueRowMatrix_[46][33] = 0.75;

      uniqueRowMatrix_[47][ 0] = 0.0234375;
      uniqueRowMatrix_[47][ 1] = -0.0390625;
      uniqueRowMatrix_[47][ 4] = -0.03125;
      uniqueRowMatrix_[47][ 5] = -0.046875;
      uniqueRowMatrix_[47][ 6] = 0.09375;
      uniqueRowMatrix_[47][ 7] = -0.0625;
      uniqueRowMatrix_[47][10] = -0.0625;
      uniqueRowMatrix_[47][13] = 0.0625;
      uniqueRowMatrix_[47][16] = 0.0625;
      uniqueRowMatrix_[47][22] = -0.125;
      uniqueRowMatrix_[47][25] = 0.375;
      uniqueRowMatrix_[47][28] = 0.1875;
      uniqueRowMatrix_[47][29] = -0.1875;
      uniqueRowMatrix_[47][31] = 0.1875;
      uniqueRowMatrix_[47][32] = -0.1875;
      uniqueRowMatrix_[47][34] = 0.75;

      uniqueRowMatrix_[48][ 0] = -0.0390625;
      uniqueRowMatrix_[48][ 1] = 0.2734375;
      uniqueRowMatrix_[48][ 4] = 0.21875;
      uniqueRowMatrix_[48][ 5] = -0.546875;
      uniqueRowMatrix_[48][ 6] = 1.09375;

      uniqueRowMatrix_[49][ 0] = 0.0234375;
      uniqueRowMatrix_[49][ 1] = -0.0390625;
      uniqueRowMatrix_[49][ 4] = -0.15625;
      uniqueRowMatrix_[49][ 5] = 0.703125;
      uniqueRowMatrix_[49][ 6] = 0.46875;

      uniqueRowMatrix_[50][ 0] = -0.0390625;
      uniqueRowMatrix_[50][ 1] = -0.0390625;
      uniqueRowMatrix_[50][ 4] = 0.15625;
      uniqueRowMatrix_[50][ 5] = -0.234375;
      uniqueRowMatrix_[50][ 6] = 0.15625;
      uniqueRowMatrix_[50][10] = 0.0625;
      uniqueRowMatrix_[50][16] = 0.3125;
      uniqueRowMatrix_[50][28] = -0.3125;
      uniqueRowMatrix_[50][29] = 0.9375;

      uniqueRowMatrix_[51][ 0] = -0.0390625;
      uniqueRowMatrix_[51][ 1] = 0.0234375;
      uniqueRowMatrix_[51][ 4] = 0.09375;
      uniqueRowMatrix_[51][ 5] = -0.046875;
      uniqueRowMatrix_[51][ 6] = -0.03125;
      uniqueRowMatrix_[51][10] = 0.125;
      uniqueRowMatrix_[51][11] = -0.125;
      uniqueRowMatrix_[51][16] = -0.125;
      uniqueRowMatrix_[51][17] = 0.375;
      uniqueRowMatrix_[51][28] = -0.375;
      uniqueRowMatrix_[51][29] = 0.375;
      uniqueRowMatrix_[51][30] = 0.75;

      uniqueRowMatrix_[52][ 0] = -0.0390625;
      uniqueRowMatrix_[52][ 1] = -0.0390625;
      uniqueRowMatrix_[52][ 4] = 0.15625;
      uniqueRowMatrix_[52][ 5] = -0.234375;
      uniqueRowMatrix_[52][ 6] = 0.15625;
      uniqueRowMatrix_[52][ 7] = 0.0625;
      uniqueRowMatrix_[52][13] = 0.3125;
      uniqueRowMatrix_[52][31] = -0.3125;
      uniqueRowMatrix_[52][32] = 0.9375;

      uniqueRowMatrix_[53][ 0] = -0.0390625;
      uniqueRowMatrix_[53][ 1] = 0.0234375;
      uniqueRowMatrix_[53][ 4] = 0.09375;
      uniqueRowMatrix_[53][ 5] = -0.046875;
      uniqueRowMatrix_[53][ 6] = -0.03125;
      uniqueRowMatrix_[53][ 7] = 0.125;
      uniqueRowMatrix_[53][ 8] = -0.125;
      uniqueRowMatrix_[53][13] = -0.125;
      uniqueRowMatrix_[53][14] = 0.375;
      uniqueRowMatrix_[53][31] = -0.375;
      uniqueRowMatrix_[53][32] = 0.375;
      uniqueRowMatrix_[53][33] = 0.75;

      uniqueRowMatrix_[54][ 0] = -0.0390625;
      uniqueRowMatrix_[54][ 1] = 0.0234375;
      uniqueRowMatrix_[54][ 4] = 0.09375;
      uniqueRowMatrix_[54][ 5] = -0.046875;
      uniqueRowMatrix_[54][ 6] = -0.03125;
      uniqueRowMatrix_[54][ 7] = 0.0625;
      uniqueRowMatrix_[54][10] = 0.0625;
      uniqueRowMatrix_[54][13] = -0.0625;
      uniqueRowMatrix_[54][16] = -0.0625;
      uniqueRowMatrix_[54][22] = 0.375;
      uniqueRowMatrix_[54][25] = -0.125;
      uniqueRowMatrix_[54][28] = -0.1875;
      uniqueRowMatrix_[54][29] = 0.1875;
      uniqueRowMatrix_[54][31] = -0.1875;
      uniqueRowMatrix_[54][32] = 0.1875;
      uniqueRowMatrix_[54][34] = 0.75;


      rowDOFPtrs_[ 0] = parentDOFs_ +  0;
      rowDOFPtrs_[ 1] = parentDOFs_ +  1;
      rowDOFPtrs_[ 2] = parentDOFs_ +  2;
      rowDOFPtrs_[ 3] = parentDOFs_ +  3;
      rowDOFPtrs_[ 4] = child0DOFs_ + 11;
      rowDOFPtrs_[ 5] = child0DOFs_ +  3;
      rowDOFPtrs_[ 6] = child1DOFs_ + 11;
      rowDOFPtrs_[ 7] = parentDOFs_ +  7;
      rowDOFPtrs_[ 8] = parentDOFs_ +  8;
      rowDOFPtrs_[ 9] = parentDOFs_ +  9;
      rowDOFPtrs_[10] = parentDOFs_ + 10;
      rowDOFPtrs_[11] = parentDOFs_ + 11;
      rowDOFPtrs_[12] = parentDOFs_ + 12;
      rowDOFPtrs_[13] = parentDOFs_ + 13;
      rowDOFPtrs_[14] = parentDOFs_ + 14;
      rowDOFPtrs_[15] = parentDOFs_ + 15;
      rowDOFPtrs_[16] = parentDOFs_ + 16;
      rowDOFPtrs_[17] = parentDOFs_ + 17;
      rowDOFPtrs_[18] = parentDOFs_ + 18;
      rowDOFPtrs_[19] = parentDOFs_ + 19;
      rowDOFPtrs_[20] = parentDOFs_ + 20;
      rowDOFPtrs_[21] = parentDOFs_ + 21;
      rowDOFPtrs_[22] = parentDOFs_ + 22;
      rowDOFPtrs_[23] = parentDOFs_ + 23;
      rowDOFPtrs_[24] = parentDOFs_ + 24;
      rowDOFPtrs_[25] = parentDOFs_ + 25;
      rowDOFPtrs_[26] = parentDOFs_ + 26;
      rowDOFPtrs_[27] = parentDOFs_ + 27;
      rowDOFPtrs_[28] = child0DOFs_ + 27;
      rowDOFPtrs_[29] = child1DOFs_ + 27;
      rowDOFPtrs_[30] = child0DOFs_ + 20;
      rowDOFPtrs_[31] = child0DOFs_ + 30;
      rowDOFPtrs_[32] = child1DOFs_ + 30;
      rowDOFPtrs_[33] = child0DOFs_ + 17;
      rowDOFPtrs_[34] = child0DOFs_ + 24;
      rowDOFPtrs_[35] = child0DOFs_ + 10;
      rowDOFPtrs_[36] = child0DOFs_ + 12;
      rowDOFPtrs_[37] = child0DOFs_ + 16;
      rowDOFPtrs_[38] = child0DOFs_ + 18;
      rowDOFPtrs_[39] = child0DOFs_ + 19;
      rowDOFPtrs_[40] = child0DOFs_ + 21;
      rowDOFPtrs_[41] = child0DOFs_ + 22;
      rowDOFPtrs_[42] = child0DOFs_ + 23;
      rowDOFPtrs_[43] = child0DOFs_ + 25;
      rowDOFPtrs_[44] = child0DOFs_ + 26;
      rowDOFPtrs_[45] = child0DOFs_ + 28;
      rowDOFPtrs_[46] = child0DOFs_ + 29;
      rowDOFPtrs_[47] = child0DOFs_ + 34;
      rowDOFPtrs_[48] = child1DOFs_ + 10;
      rowDOFPtrs_[49] = child1DOFs_ + 12;
      rowDOFPtrs_[50] = child1DOFs_ + 25;
      rowDOFPtrs_[51] = child1DOFs_ + 26;
      rowDOFPtrs_[52] = child1DOFs_ + 28;
      rowDOFPtrs_[53] = child1DOFs_ + 29;
      rowDOFPtrs_[54] = child1DOFs_ + 34;

      colDOFPtrs_[ 0] = parentDOFs_ +  0;
      colDOFPtrs_[ 1] = parentDOFs_ +  1;
      colDOFPtrs_[ 2] = parentDOFs_ +  2;
      colDOFPtrs_[ 3] = parentDOFs_ +  3;
      colDOFPtrs_[ 4] = parentDOFs_ +  4;
      colDOFPtrs_[ 5] = parentDOFs_ +  5;
      colDOFPtrs_[ 6] = parentDOFs_ +  6;
      colDOFPtrs_[ 7] = parentDOFs_ +  7;
      colDOFPtrs_[ 8] = parentDOFs_ +  8;
      colDOFPtrs_[ 9] = parentDOFs_ +  9;
      colDOFPtrs_[10] = parentDOFs_ + 10;
      colDOFPtrs_[11] = parentDOFs_ + 11;
      colDOFPtrs_[12] = parentDOFs_ + 12;
      colDOFPtrs_[13] = parentDOFs_ + 13;
      colDOFPtrs_[14] = parentDOFs_ + 14;
      colDOFPtrs_[15] = parentDOFs_ + 15;
      colDOFPtrs_[16] = parentDOFs_ + 16;
      colDOFPtrs_[17] = parentDOFs_ + 17;
      colDOFPtrs_[18] = parentDOFs_ + 18;
      colDOFPtrs_[19] = parentDOFs_ + 19;
      colDOFPtrs_[20] = parentDOFs_ + 20;
      colDOFPtrs_[21] = parentDOFs_ + 21;
      colDOFPtrs_[22] = parentDOFs_ + 22;
      colDOFPtrs_[23] = parentDOFs_ + 23;
      colDOFPtrs_[24] = parentDOFs_ + 24;
      colDOFPtrs_[25] = parentDOFs_ + 25;
      colDOFPtrs_[26] = parentDOFs_ + 26;
      colDOFPtrs_[27] = parentDOFs_ + 27;
      colDOFPtrs_[28] = parentDOFs_ + 28;
      colDOFPtrs_[29] = parentDOFs_ + 29;
      colDOFPtrs_[30] = parentDOFs_ + 30;
      colDOFPtrs_[31] = parentDOFs_ + 31;
      colDOFPtrs_[32] = parentDOFs_ + 32;
      colDOFPtrs_[33] = parentDOFs_ + 33;
      colDOFPtrs_[34] = parentDOFs_ + 34;

      type0_rowDOFPtrs_[ 0] = parentDOFs_ +  0;
      type0_rowDOFPtrs_[ 1] = parentDOFs_ +  1;
      type0_rowDOFPtrs_[ 2] = parentDOFs_ +  2;
      type0_rowDOFPtrs_[ 3] = parentDOFs_ +  3;
      type0_rowDOFPtrs_[ 4] = child0DOFs_ + 11;
      type0_rowDOFPtrs_[ 5] = child0DOFs_ +  3;
      type0_rowDOFPtrs_[ 6] = child1DOFs_ + 11;
      type0_rowDOFPtrs_[ 7] = parentDOFs_ +  7;
      type0_rowDOFPtrs_[ 8] = parentDOFs_ +  8;
      type0_rowDOFPtrs_[ 9] = parentDOFs_ +  9;
      type0_rowDOFPtrs_[10] = parentDOFs_ + 10;
      type0_rowDOFPtrs_[11] = parentDOFs_ + 11;
      type0_rowDOFPtrs_[12] = parentDOFs_ + 12;
      type0_rowDOFPtrs_[13] = parentDOFs_ + 13;
      type0_rowDOFPtrs_[14] = parentDOFs_ + 14;
      type0_rowDOFPtrs_[15] = parentDOFs_ + 15;
      type0_rowDOFPtrs_[16] = parentDOFs_ + 16;
      type0_rowDOFPtrs_[17] = parentDOFs_ + 17;
      type0_rowDOFPtrs_[18] = parentDOFs_ + 18;
      type0_rowDOFPtrs_[19] = parentDOFs_ + 19;
      type0_rowDOFPtrs_[20] = parentDOFs_ + 20;
      type0_rowDOFPtrs_[21] = parentDOFs_ + 21;
      type0_rowDOFPtrs_[22] = parentDOFs_ + 22;
      type0_rowDOFPtrs_[23] = parentDOFs_ + 23;
      type0_rowDOFPtrs_[24] = parentDOFs_ + 24;
      type0_rowDOFPtrs_[25] = parentDOFs_ + 25;
      type0_rowDOFPtrs_[26] = parentDOFs_ + 26;
      type0_rowDOFPtrs_[27] = parentDOFs_ + 27;
      type0_rowDOFPtrs_[28] = child0DOFs_ + 27;
      type0_rowDOFPtrs_[29] = child1DOFs_ + 30;
      type0_rowDOFPtrs_[30] = child0DOFs_ + 20;
      type0_rowDOFPtrs_[31] = child0DOFs_ + 30;
      type0_rowDOFPtrs_[32] = child1DOFs_ + 27;
      type0_rowDOFPtrs_[33] = child0DOFs_ + 17;
      type0_rowDOFPtrs_[34] = child0DOFs_ + 24;
      type0_rowDOFPtrs_[35] = child0DOFs_ + 10;
      type0_rowDOFPtrs_[36] = child0DOFs_ + 12;
      type0_rowDOFPtrs_[37] = child0DOFs_ + 16;
      type0_rowDOFPtrs_[38] = child0DOFs_ + 18;
      type0_rowDOFPtrs_[39] = child0DOFs_ + 19;
      type0_rowDOFPtrs_[40] = child0DOFs_ + 21;
      type0_rowDOFPtrs_[41] = child0DOFs_ + 22;
      type0_rowDOFPtrs_[42] = child0DOFs_ + 23;
      type0_rowDOFPtrs_[43] = child0DOFs_ + 25;
      type0_rowDOFPtrs_[44] = child0DOFs_ + 26;
      type0_rowDOFPtrs_[45] = child0DOFs_ + 28;
      type0_rowDOFPtrs_[46] = child0DOFs_ + 29;
      type0_rowDOFPtrs_[47] = child0DOFs_ + 34;
      type0_rowDOFPtrs_[48] = child1DOFs_ + 10;
      type0_rowDOFPtrs_[49] = child1DOFs_ + 12;
      type0_rowDOFPtrs_[50] = child1DOFs_ + 28;
      type0_rowDOFPtrs_[51] = child1DOFs_ + 29;
      type0_rowDOFPtrs_[52] = child1DOFs_ + 25;
      type0_rowDOFPtrs_[53] = child1DOFs_ + 26;
      type0_rowDOFPtrs_[54] = child1DOFs_ + 34;

      type0_colDOFPtrs_[ 0] = parentDOFs_ +  0;
      type0_colDOFPtrs_[ 1] = parentDOFs_ +  1;
      type0_colDOFPtrs_[ 2] = parentDOFs_ +  2;
      type0_colDOFPtrs_[ 3] = parentDOFs_ +  3;
      type0_colDOFPtrs_[ 4] = parentDOFs_ +  4;
      type0_colDOFPtrs_[ 5] = parentDOFs_ +  5;
      type0_colDOFPtrs_[ 6] = parentDOFs_ +  6;
      type0_colDOFPtrs_[ 7] = parentDOFs_ +  7;
      type0_colDOFPtrs_[ 8] = parentDOFs_ +  8;
      type0_colDOFPtrs_[ 9] = parentDOFs_ +  9;
      type0_colDOFPtrs_[10] = parentDOFs_ + 10;
      type0_colDOFPtrs_[11] = parentDOFs_ + 11;
      type0_colDOFPtrs_[12] = parentDOFs_ + 12;
      type0_colDOFPtrs_[13] = parentDOFs_ + 13;
      type0_colDOFPtrs_[14] = parentDOFs_ + 14;
      type0_colDOFPtrs_[15] = parentDOFs_ + 15;
      type0_colDOFPtrs_[16] = parentDOFs_ + 16;
      type0_colDOFPtrs_[17] = parentDOFs_ + 17;
      type0_colDOFPtrs_[18] = parentDOFs_ + 18;
      type0_colDOFPtrs_[19] = parentDOFs_ + 19;
      type0_colDOFPtrs_[20] = parentDOFs_ + 20;
      type0_colDOFPtrs_[21] = parentDOFs_ + 21;
      type0_colDOFPtrs_[22] = parentDOFs_ + 22;
      type0_colDOFPtrs_[23] = parentDOFs_ + 23;
      type0_colDOFPtrs_[24] = parentDOFs_ + 24;
      type0_colDOFPtrs_[25] = parentDOFs_ + 25;
      type0_colDOFPtrs_[26] = parentDOFs_ + 26;
      type0_colDOFPtrs_[27] = parentDOFs_ + 27;
      type0_colDOFPtrs_[28] = parentDOFs_ + 28;
      type0_colDOFPtrs_[29] = parentDOFs_ + 29;
      type0_colDOFPtrs_[30] = parentDOFs_ + 30;
      type0_colDOFPtrs_[31] = parentDOFs_ + 31;
      type0_colDOFPtrs_[32] = parentDOFs_ + 32;
      type0_colDOFPtrs_[33] = parentDOFs_ + 33;
      type0_colDOFPtrs_[34] = parentDOFs_ + 34;
    }
  };
}

#endif
